local Object = require "core.Object"
local DamageUtil = require "util.DamageUtil"
local Result = require "game.Result"

local Action = Object:extend()

Action.SKILL_ATTACK_ID = 1
Action.SKILL_DEFEND_ID = 2

Action.PRIORITY_HIGHEST = 0
Action.PRIORITY_HIGH = 1
Action.PRIORITY_NORMAL = 2
Action.PRIORITY_LOW = 3

Action.MOTION_NONE = 0
Action.MOTION_ATTACK = 1
Action.MOTION_CUSTOM = 2
Action.MOTION_CAST = 3
Action.MOTION_ITEM = 4

Action.ANIMATION_ITEM_ID = 1

function Action:__lt(a)
	return self:getPriority() > a:getPriority() or (self:getPriority() == a:getPriority() and self.user:getSpd() < a.user:getSpd())
end

function Action:initialize(code, id, user, targetAll, targetGroup, targetIndex)
    self.errors = {}
    self.messages = {}
    self.targets = {}
    self.results = {}
    self.code = code
    self.id = id
    self.user = user
    self.targetAll = targetAll
    self.targetGroup = targetGroup
    self.targetIndex = targetIndex
end

function Action:isNone()
    return self.code == "None"
end

function Action:isSkill()
    return self.code == "Skill"
end

function Action:isItem()
    return self.code == "Item"
end

function Action:isAttack()
    return self:isSkill() and self.id == Action.SKILL_ATTACK_ID
end

function Action:isDefend()
    return self:isSkill() and self.id == Action.SKILL_DEFEND_ID
end

function Action:isSystem()
    return self:isSkill() and dataSkills[self.id].type == 1
end

function Action:isTechnique()
    return self:isSkill() and dataSkills[self.id].type == 2
end

function Action:isMagic()
    return self:isSkill() and dataSkills[self.id].type == 3
end

function Action:isSummon()
    return self:isSkill() and dataSkills[self.id].type == 4
end

function Action:isValidTarget(target)
    return not target:isDead() or (target:isActor() and self:isDeadOnly())
end

function Action:isDeadOnly()
    return (self:isSkill() and dataSkills[self.id].deadOnly) or (self:isItem() and dataItems[self.id].deadOnly)
end

function Action:isPhysical()
    return self:isSkill() and dataSkills[self.id].physical
end

function Action:isHpAltered()
    return self:getHpDamageType() ~= 0
end

function Action:isMpAltered()
    return self:getMpDamageType() ~= 0
end

function Action:hasErrors()
    return #self.errors > 0
end

function Action:hasName()
    return not self:isSystem() or (self.id == Action.SKILL_DEFEND_ID and self.user:isEnemy())
end

function Action:hasAttackMotion()
    return self:getMotionType() == Action.MOTION_ATTACK
end

function Action:hasCustomMotion()
    return self:getMotionType() == Action.MOTION_CUSTOM
end

function Action:hasCastMotion()
    return self:getMotionType() == Action.MOTION_CAST
end

function Action:hasItemMotion()
    return self:getMotionType() == Action.MOTION_ITEM
end

function Action:hasReturnMotion()
    return self:hasAttackMotion() or self:hasCustomMotion() or self:hasCastMotion() or self:hasItemMotion()
end

function Action:hasUserAnimation()
    return self:getUserAnimation() ~= 0
end

function Action:hasAnimation()
    return self:getAnimation() ~= 0
end

function Action:hasCriticals()
    for _, v in ipairs(self.results) do
        if v:isSuccess() and v.critical then
            return true
        end
    end
    return false
end

function Action:hasHpDrainEffect()
    return (self:isSkill() and dataSkills[self.id].hpDrain) or (self:isItem() and dataItems[self.id].hpDrain)
end

function Action:hasMpDrainEffect()
    return (self:isSkill() and dataSkills[self.id].mpDrain) or (self:isItem() and dataItems[self.id].mpDrain)
end

function Action:getErrors()
    return self.errors
end

function Action:getMessages()
    return self.messages
end

function Action:getTargets()
    return self.targets
end

function Action:getResults()
    return self.results
end

function Action:getUser()
    return self.user
end

function Action:getPriority()
    return (self:isSkill() and dataSkills[self.id].priority) or Action.PRIORITY_NORMAL
end

function Action:getName()
    return (self:isSkill() and dataSkills[self.id].name) or (self:isItem() and dataItems[self.id].name) or ""
end

function Action:getMotionType()
    return (self:isSkill() and dataSkills[self.id].motionType) or (self:isItem() and Action.MOTION_ITEM) or Action.MOTION_NONE
end

function Action:getMotion()
    return (self:isSkill() and dataSkills[self.id].motion) or 0
end

function Action:getUserAnimation()
    return (self:isSkill() and dataSkills[self.id].userAnimation) or (self:isItem() and Action.ANIMATION_ITEM_ID) or 0
end

function Action:getAnimation()
    local id = (self:isSkill() and dataSkills[self.id].animation) or (self:isItem() and dataItems[self.id].animation) or 0
    return (self:isPhysical() and id == 0 and self.user:getAttackAnimation()) or id
end

function Action:getSuccessRate()
    return (self:isSkill() and dataSkills[self.id].successRate) or 1
end

function Action:getHpDamageType()
    return (self:isSkill() and dataSkills[self.id].hpDamageType) or (self:isItem() and dataItems[self.id].hpDamageType) or 0
end

function Action:getHpDamageFormula()
    return (self:isSkill() and dataSkills[self.id].hpDamageFormula) or (self:isItem() and dataItems[self.id].hpDamageFormula) or ""
end

function Action:getHpDamageVariance()
    return (self:isSkill() and dataSkills[self.id].hpDamageVariance) or (self:isItem() and dataItems[self.id].hpDamageVariance) or 0
end

function Action:getHpDrainRate()
    return (self:isSkill() and dataSkills[self.id].hpDrainRate) or (self:isItem() and dataItems[self.id].hpDrainRate) or 0
end

function Action:getMpDamageType()
    return (self:isSkill() and dataSkills[self.id].mpDamageType) or (self:isItem() and dataItems[self.id].mpDamageType) or 0
end

function Action:getMpDamageFormula()
    return (self:isSkill() and dataSkills[self.id].mpDamageFormula) or (self:isItem() and dataItems[self.id].mpDamageFormula) or ""
end

function Action:getMpDamageVariance()
    return (self:isSkill() and dataSkills[self.id].mpDamageVariance) or (self:isItem() and dataItems[self.id].mpDamageVariance) or 0
end

function Action:getMpDrainRate()
    return (self:isSkill() and dataSkills[self.id].mpDrainRate) or (self:isItem() and dataItems[self.id].mpDrainRate) or 0
end

function Action:getAttributes()
    local r = {unpack((self:isSkill() and dataSkills[self.id].attributes) or (self:isItem() and dataItems[self.id].attributes) or {})}
    if self:isPhysical() then
        for _, v1 in ipairs(self.user:getPhysicalAttributes()) do
            local found = false
            for _, v2 in ipairs(r) do
                if v2 == v1 then
                    found = true
                    break
                end
            end
            if not found then
                table.insert(r, v1)
            end
        end
    end
    return r
end

function Action:getAddableStatuses()
    local r = {unpack((self:isSkill() and dataSkills[self.id].addStatuses) or (self:isItem() and dataItems[self.id].addStatuses) or {})}
    if self:isPhysical() then
        for _, v in ipairs(self.user:getPhysicalStatuses()) do
            table.insert(r, v)
        end
    end
    return r
end

function Action:getAddableBuffs()
    return (self:isSkill() and dataSkills[self.id].addBuffs) or (self:isItem() and dataItems[self.id].addBuffs) or {}
end

function Action:getAddableDebuffs()
    return (self:isSkill() and dataSkills[self.id].addDebuffs) or (self:isItem() and dataItems[self.id].addDebuffs) or {}
end

function Action:getRemovableStatuses()
    return (self:isSkill() and dataSkills[self.id].removeStatuses) or (self:isItem() and dataItems[self.id].removeStatuses) or {}
end

function Action:makeNone()
    self.code = "None"
    self.id = 0
    self.targetAll = false
    self.targetGroup = ""
    self.targetIndex = 0
end

function Action:makeRestrictedAttack()
    self.code = "Skill"
    self.id = Action.SKILL_ATTACK_ID
    self.targetAll = false
    self.targetGroup = ""
    self.targetIndex = 0
    local r = self.user:getRestriction()
    if r == 1 then
        self.targetGroup = (self.user:isActor() and "E") or (self.user:isEnemy() and "A") or ""
    end
    if r == 2 then
        self.targetGroup = (self.user:isActor() and "A") or (self.user:isEnemy() and "E") or ""
    end
    if r == 3 then
        self.targetGroup = ({"A", "E"})[math.random(1, 2)]
    end
end

function Action:checkErrors()
    if self:isSkill() then
        if self.user:getMp() < dataSkills[self.id].mpCost then
            table.insert(self.errors, TM.BATTLE_ERROR_SKILL_MP)
        end
        if self.user:isSkillLocked(dataSkills[self.id].type) then
            table.insert(self.errors, TM.BATTLE_ERROR_SKILL_LOCKED)
        end
    end
    if self:isItem() and gameActorGroup:getItemAmount(self.id) < 1 then
        table.insert(self.errors, TM.BATTLE_ERROR_ITEM_AMOUNT)
    end
end

function Action:setupIndex()
    if not self.targetAll then
        local group = (self.targetGroup == "A" and gameActorGroup) or (self.targetGroup == "E" and gameEnemyGroup) or nil
        if self.targetIndex == 0 then
            -- Random index
            local indexes = {}
            for i, v in ipairs(group:getBattleMembers()) do
                if self:isValidTarget(v) then
                    table.insert(indexes, i)
                end
            end
            self.targetIndex = indexes[math.random(1, #indexes)]
        else
            -- Adjust index
            local index = self.targetIndex
            while not self:isValidTarget(group:getMember(self.targetIndex)) do
                self.targetIndex = self.targetIndex % group:getBattleSize() + 1
                if self.targetIndex == index then
                    break
                end
            end
        end
    end
end

function Action:setupTargets()
    local group = (self.targetGroup == "A" and gameActorGroup) or (self.targetGroup == "E" and gameEnemyGroup) or nil
    if self.targetAll then
        for _, v in ipairs(group:getBattleMembers()) do
            if self:isValidTarget(v) then
                table.insert(self.targets, v)
            end
        end
    else
        table.insert(self.targets, group:getMember(self.targetIndex))
    end
end

function Action:setupResults()
    for _, v in ipairs(self.targets) do
        local r = Result:new()
        -- Target
        r.target = v
        -- Fail/miss/evade
        r.failed = (self:isDeadOnly() and not v:isDead()) or math.random() > self:getSuccessRate()
        r.missed = (self:isPhysical() and math.random() > self.user:getHit())
        r.evaded = (self:isPhysical() and math.random() <= v:getEva())
        -- Critical
        r.critical = (self:isPhysical() and math.random() <= self.user:getCrt())
        -- Hp damage
        if self:isHpAltered() then
            r.alterHp = true
            r.hpValue = DamageUtil:calcValue(self.user, v, self:getHpDamageFormula(), self:getHpDamageVariance(), self:getHpDamageType(), r.critical, self:getAttributes())
            if self:hasHpDrainEffect() then
                r.alterUserHp = true
                r.userHpValue = DamageUtil:calcDrain(self.user, r.hpValue, self:getHpDrainRate())
            end
        end
        -- Mp damage
        if self:isMpAltered() then
            r.alterMp = true
            r.mpValue = DamageUtil:calcValue(self.user, v, self:getMpDamageFormula(), self:getMpDamageVariance(), self:getMpDamageType(), r.critical, self:getAttributes())
            if self:hasMpDrainEffect() then
                r.alterUserMp = true
                r.userMpValue = DamageUtil:calcDrain(self.user, r.mpValue, self:getMpDrainRate())
            end
        end
        -- Statuses/buffs/debuffs
        for _, s in ipairs(self:getAddableStatuses()) do
            if math.random() <= s.rate * v:getStatusRate(s.id) then
                table.insert(r.addedStatuses, s.id)
            end
        end
        for _, b in ipairs(self:getAddableBuffs()) do
            table.insert(r.addedBuffs, b)
        end
        for _, d in ipairs(self:getAddableDebuffs()) do
            if math.random() <= d.rate * v:getDebuffRate(d.id) then
                table.insert(r.addedDebuffs, d)
            end
        end
        for _, s in ipairs(self:getRemovableStatuses()) do
            table.insert(r.removedStatuses, s)
        end
        -- Result
        table.insert(self.results, r)
    end
end

function Action:setup()
    if self.user:isRestricted() then
        if self.user:isConscious() then
            self:makeRestrictedAttack()
        else
            self:makeNone()
        end
    end
    if not self:isNone() then
        self:checkErrors()
        if not self:hasErrors() then
            self:setupIndex()
            self:setupTargets()
            self:setupResults()
        end
    end
end

function Action:applyCosts()
    if self:isSkill() then
        self.user:changeMp(-dataSkills[self.id].mpCost)
    end
    if self:isItem() then
        if self.user:isActor() then
            gameActorGroup:changeItemAmount(self.id, -1)
        end
    end
end

function Action:applyResults()
    local userHpValue = 0
    local userMpValue = 0
    for _, v in ipairs(self.results) do
        if v:isSuccess() then
            -- Target
            for _, s in ipairs(v.addedStatuses) do
                v.target:addStatus(s)
            end
            for _, b in ipairs(v.addedBuffs) do
                v.target:addBuff(b.id, b.turns)
            end
            for _, d in ipairs(v.addedDebuffs) do
                v.target:addDebuff(d.id, d.turns)
            end
            for _, s in ipairs(v.removedStatuses) do
                v.target:removeStatus(s)
            end
            if self:isPhysical() and v.alterHp and v.hpValue < 0 then
                v.target:removeStatusesOnDamage()
            end
            v.target:changeHp(v.hpValue)
            v.target:changeMp(v.mpValue)
            -- User
            userHpValue = userHpValue + v.userHpValue
            userMpValue = userMpValue + v.userMpValue
        end
    end
    self.user:changeHp(userHpValue)
    self.user:changeMp(userMpValue)
end

return Action