local Drawable = require "game.Drawable"

local Animation = Drawable:extend()

Animation.POSITION_TARGET_TOP = 1
Animation.POSITION_TARGET_CENTER = 2
Animation.POSITION_TARGET_BOTTOM = 3
Animation.POSITION_GLOBAL = 4
Animation.POSITION_SCREEN = 5

function Animation:initialize(id, targets)
	-- Shader
	self.shader = RM:getShader("default.glsl")
	-- Image/quads
	self.image = RM:getAnimationImage(dataAnimations[id].image)
	self.quads = {}
	local qw = self:getQuadW()
	local qh = self:getQuadH()
	local iw = qw * 8
	local ih = qh * 4
	for y = 0, 3 do
		for x = 0, 7 do
			local qx = x * qw
			local qy = y * qh
			table.insert(self.quads, love.graphics.newQuad(qx, qy, qw, qh, iw, ih))
		end
	end
	-- Id/state
	self.id = id
	self.targets = targets
	self.frame = 1
	self.timer = 0
end

function Animation:isBusy()
	return self.frame <= #dataAnimations[self.id].frames
end

function Animation:getQuadW()
	return 192
end

function Animation:getQuadH()
	return 192
end

function Animation:getZ()
	return Animation.PRIORITY_ANIMATION
end

function Animation:update()
	if self.frame <= #dataAnimations[self.id].frames then
		if self.timer == 0 then
			local frame = dataAnimations[self.id].frames[self.frame]
			local sound = frame.sound
			local flash = frame.flash
			local flashDisplay = frame.flashDisplay
			if sound then
				AM:playSound(sound)
			end
			if flash then
				for _, v in ipairs(self.targets) do
					v:flash(unpack(flash))
				end
			end
			if flashDisplay then
				gameDisplay:flash(unpack(flashDisplay))
			end
		end
		self.timer = self.timer + 1
		if self.timer >= 128 / (2 ^ dataAnimations[self.id].speed) then
			self.frame = self.frame + 1
			self.timer = 0
		end
	end
end

function Animation:drawQuads(x, y)
	for _, v in ipairs(dataAnimations[self.id].frames[self.frame].quads or {}) do
		local index = v[1]
		local x = x + v[2]
		local y = y + v[3]
		local z = v[4]
		local r = v[5]
		local g = v[6]
		local b = v[7]
		local s = v[8]
		local a = v[9]
		local ox = self:getQuadW() / 2
		local oy = self:getQuadH() / 2
		self.shader:send("opacity", a)
		self.shader:sendColor("color", {r, g, b, s})
		love.graphics.setColor(1, 1, 1)
		love.graphics.draw(self.image, self.quads[index], x, y, 0, z, z, ox, oy)
	end
end

function Animation:draw()
	if self.frame > #dataAnimations[self.id].frames then
		return
	end
	local position = dataAnimations[self.id].position
	self.shader:sendColor("flash", {1, 1, 1, 0})
	self.shader:sendColor("displayColor", {gameDisplay:getColor()})
	self.shader:sendColor("displayFlash", {gameDisplay:getFlash()})
	love.graphics.setShader(self.shader)
	if position == Animation.POSITION_TARGET_TOP or position == Animation.POSITION_TARGET_CENTER or position == Animation.POSITION_TARGET_BOTTOM then
		for _, v in ipairs(self.targets) do
			local h = v:getH()
			local x = v:getX()
			local y = v:getY()
			if position == Animation.POSITION_TARGET_TOP then
				y = y - h
			end
			if position == Animation.POSITION_TARGET_CENTER then
				y = y - h / 2
			end
			self:drawQuads(x, y)
		end
	end
	if position == Animation.POSITION_GLOBAL then
		local x = 0
		local y = 0
		for _, v in ipairs(self.targets) do
			x = x + v:getX()
			y = y + v:getY()
		end
		x = x / #self.targets
		y = y / #self.targets
		self:drawQuads(x, y)
	end
	if position == Animation.POSITION_SCREEN then
		local x = VM:getW() / 2
		local y = VM:getH() / 2
		self:drawQuads(x, y)
	end
end

return Animation