local Object = require "core.Object"
local DamageUtil = require "util.DamageUtil"
local Damage = require "game.Damage"
local Animation = require "game.Animation"

local Battle = Object:extend()

function Battle:initialize()
	self.fighters = {}
	self.messages = {}
	self.damage = {}
	self.animations = {}
	self.refreshables = {}
	self.actionStack = {}
	self.action = nil
	self.actionUser = nil
	self.actorPreemptive = false
	self.enemyPreemptive = false
	self.escapeAttempt = false
	self.escapeSuccess = false
	self.escapeRate = 0
	self.phase = ""
	self.index = 0
end

---------------------------------------------------------------------------------------------------
-- Getters/setters
---------------------------------------------------------------------------------------------------

function Battle:isBusy()
	if AM:isBusy() then
		return true
	end
	if SM:isStateChanging() then
		return true
	end
	if gameTransition:isBusy() then
		return true
	end
	if gameDisplay:isBusy() then
		return true
	end
	for _, v in ipairs(self.fighters) do
		if v:isMotionBusy() then
			return true
		end
	end
	if self:hasMessages() then
		return true
	end
	if self:hasAnimations() then
		return true
	end
	if self:hasDamage() then
		return true
	end
	return false
end

function Battle:isEscapeFailure()
	return self.escapeAttempt and not self.escapeSuccess
end

function Battle:isEscapeSuccess()
	return self.escapeSuccess
end

function Battle:isActorDefeated()
	return gameActorGroup:isDefeated()
end

function Battle:isEnemyDefeated()
	return gameEnemyGroup:isDefeated()
end

function Battle:isOver()
	return self:isEscapeSuccess() or self:isActorDefeated() or self:isEnemyDefeated()
end

function Battle:isGroupInputPhase()
	return self.phase == "Input" and self.index == 0
end

function Battle:isActorInputPhase()
	return self.phase == "Input" and self.index > 0 and self.index <= gameActorGroup:getBattleSize()
end

function Battle:hasMessages()
	return #self.messages > 0
end

function Battle:hasAnimations()
	return #self.animations > 0
end

function Battle:hasDamage()
	return #self.damage > 0
end

function Battle:hasRefreshable(refreshable)
	for _, v in ipairs(self.refreshables) do
		if v == refreshable then
			return true
		end
	end
	return false
end

function Battle:getFighters()
	return self.fighters
end

function Battle:getMessage()
	return self.messages[1]
end

function Battle:getAnimations()
	return self.animations
end

function Battle:getDamage()
	return self.damage
end

function Battle:getIndex()
	return self.index
end

---------------------------------------------------------------------------------------------------
-- Methods
---------------------------------------------------------------------------------------------------

function Battle:addMessage(type, text)
	table.insert(self.messages, {type = type, text = text})
end

function Battle:popMessage()
	table.remove(self.messages, 1)
end

function Battle:addDamage(type, value, x, y, delay)
	table.insert(self.damage, Damage:new(type, value, x, y, delay))
end

function Battle:addAnimation(id, targets)
	table.insert(self.animations, Animation:new(id, targets))
end

function Battle:addRefreshable(refreshable)
	if not self:hasRefreshable(refreshable) then
		table.insert(self.refreshables, refreshable)
	end
end

function Battle:selectEscape()
	self.escapeAttempt = true
	self.escapeSuccess = self.actorPreemptive or math.random() < self.escapeRate
	if not self.escapeSuccess then
		self.escapeRate = self.escapeRate + 0.1
	end
end

function Battle:selectNextActor()
	repeat
		self.index = self.index + 1
	until self.index > gameActorGroup:getBattleSize() or not gameActorGroup:getMember(self.index):isRestricted()
end

function Battle:selectPrevActor()
	repeat
		self.index = self.index - 1
	until self.index <= 0 or not gameActorGroup:getMember(self.index):isRestricted()
	if self.index > 0 then
		gameActorGroup:getMember(self.index):clearAction()
	end
end

function Battle:clearPreemptive()
	self.actorPreemptive = false
	self.enemyPreemptive = false
end

function Battle:clearEscape()
	self.escapeAttempt = false
	self.escapeSuccess = false
end

function Battle:setupPreemptive()
	-- TODO
end

function Battle:setupEscape()
	-- TODO
	self.escapeRate = 0.25
end

function Battle:setupFighters()
	self.fighters = {}
	for _, v in ipairs(gameActorGroup:getBattleMembers()) do
		table.insert(self.fighters, v)
	end
	for _, v in ipairs(gameEnemyGroup:getBattleMembers()) do
		table.insert(self.fighters, v)
	end
end

function Battle:setupActionStack()
	self.actionStack = {}
	for _, v in ipairs(self.fighters) do
		if (v:isActor() and not self.enemyPreemptive) or (v:isEnemy() and not self.actorPreemptive) then
			table.insert(self.actionStack, v:getAction())
		end
	end
	table.sort(self.actionStack)
end

function Battle:refreshMotions()
	for _, v in ipairs(self.refreshables) do
		v:refreshMotion()
	end
	self.refreshables = {}
end

function Battle:executeActionDamage()
	local damage = {}
	local alterUserHp = false
	local alterUserMp = false
	local userHpValue = 0
	local userMpValue = 0
	for _, v in ipairs(self.action:getResults()) do
		damage[v.target] = {}
		if not v:isSuccess() then
			table.insert(damage[v.target], {type = "Miss", value = 0})
		else
			-- TODO revive?
			if v.alterHp then
				table.insert(damage[v.target], {type = "Hp", value = v.hpValue})
			end
			if v.alterMp then
				table.insert(damage[v.target], {type = "Mp", value = v.mpValue})
			end
			if v.alterUserHp then
				alterUserHp = true
				userHpValue = userHpValue + v.userHpValue
			end
			if v.alterUserMp then
				alterUserMp = true
				userMpValue = userMpValue + v.userMpValue
			end
		end
	end
	if alterUserHp or alterUserMp then
		if not damage[self.actionUser] then
			damage[self.actionUser] = {}
		end
		if alterUserHp then
			table.insert(damage[self.actionUser], {type = "Hp", value = userHpValue})
		end
		if alterUserMp then
			table.insert(damage[self.actionUser], {type = "Mp", value = userMpValue})
		end
	end
	for k, v1 in pairs(damage) do
		for i, v2 in ipairs(v1) do
			self:addDamage(v2.type, v2.value, k:getX() - (#v1 - 1) * 4 + (i - 1) * 8, k:getY(), (i - 1) * 16)
		end
	end
end

function Battle:executeStatusDamage()
	local damage = {}
	for _, v in ipairs(self.fighters) do
		if not v:isDead() then
			local hpValue = DamageUtil:calcHpRecovery(v)
			local mpValue = DamageUtil:calcMpRecovery(v)
			v:changeHp(hpValue)
			v:changeMp(mpValue)
			if hpValue ~= 0 or mpValue ~= 0 then
				damage[v] = {}
				if hpValue ~= 0 then
					if hpValue < 0 then
						v:performMotion("Damage")
					end
					table.insert(damage[v], {type = "Hp", value = hpValue})
				end
				if mpValue ~= 0 then
					table.insert(damage[v], {type = "Mp", value = mpValue})
				end
				self:addRefreshable(v)
			end
		end
	end
	for k, v1 in pairs(damage) do
		for i, v2 in ipairs(v1) do
			self:addDamage(v2.type, v2.value, k:getX() - (#v1 - 1) * 4 + (i - 1) * 8, k:getY(), (i - 1) * 16)
		end
	end
end

function Battle:executeEscape()
	for _, v in ipairs(gameActorGroup:getBattleMembers()) do
		v:performMotion("Escape")
	end
	self:addMessage("LongAuto", TM.BATTLE_ESCAPE_SUCCESS)
	AM:playEscapeSound()
end

function Battle:executeActorDefeat()
	-- TODO musica sconfitta???
	self:addMessage("LongAuto", TM.BATTLE_ACTOR_DEFEATED)
end

function Battle:executeEnemyDefeat()
	for _, v in ipairs(gameActorGroup:getBattleMembers()) do
		v:performMotion("Win")
	end
	-- TODO
	self:addMessage("Long", TM.BATTLE_VICTORY)
	self:addMessage("Long", string.format(TM.BATTLE_EXP_OBTAINED, 0))
	self:addMessage("Long", string.format(TM.BATTLE_AP_OBTAINED, 0))
	self:addMessage("Long", string.format(TM.BATTLE_LEVEL_GAINED, "Nessuno"))
	self:addMessage("Long", string.format(TM.BATTLE_SKILL_LEARNED, "Nessuno", "Niente"))
	self:addMessage("Long", string.format(TM.BATTLE_ITEM_OBTAINED, "Nulla", 0))
	-- TODO
	AM:playVictoryMusic()
end

---------------------------------------------------------------------------------------------------
-- Start methods
---------------------------------------------------------------------------------------------------

function Battle:startBegin()
	self.phase = "Begin"
	self:clearPreemptive()
	self:clearEscape()
	self:setupPreemptive()
	self:setupEscape()
	self:setupFighters()
	gameActorGroup:setupBattlePosition()
	gameEnemyGroup:setupBattlePosition()
	for _, v in ipairs(self.fighters) do
		v:performMotion("Appear")
	end
	-- TODO
	AM:playBattleMusic()
end

function Battle:startPostBegin()
	self.phase = "PostBegin"
	if self.actorPreemptive then
		self:addMessage("LongAuto", TM.BATTLE_ACTOR_PREEMPTIVE)
	end
	if self.enemyPreemptive then
		self:addMessage("LongAuto", TM.BATTLE_ENEMY_PREEMPTIVE)
	end
end

function Battle:startInput()
	self.phase = "Input"
	self.index = 0
	for _, v in ipairs(self.fighters) do
		v:setupAction()
	end
	if self.enemyPreemptive or gameActorGroup:isRestricted() then
		self:startTurnBegin()
	end
end

function Battle:startTurnBegin()
	self.phase = "TurnBegin"
	self:setupActionStack()
	if self:isEscapeFailure() then
		self:addMessage("LongAuto", TM.BATTLE_ESCAPE_FAILURE)
	end
end

function Battle:startActionBegin()
	self.phase = "ActionBegin"
	self.action = table.remove(self.actionStack)
	self.actionUser = self.action:getUser()
	self.action:setup()
	if not self.action:isNone() then
		if self.action:hasName() then
			self:addMessage("Auto", self.action:getName())
		end
		if self.action:hasErrors() then
			for _, v in ipairs(self.action:getErrors()) do
				self:addMessage("LongAuto", v)
			end
		end
	end
end

function Battle:startActionProcess1()
	self.phase = "ActionProcess1"
	if self.action:hasAttackMotion() then
		self.actionUser:performMotion("Attack")
	end
	if self.action:hasCustomMotion() then
		self.actionUser:performMotion("Custom", self.action:getMotion())
	end
	if self.action:hasCastMotion() then
		self.actionUser:performMotion("Cast")
	end
	if self.action:hasItemMotion() then
		self.actionUser:performMotion("Item")
	end
end

function Battle:startActionProcess2()
	self.phase = "ActionProcess2"
	self.action:applyCosts()
	if self.action:hasUserAnimation() then
		self:addAnimation(self.action:getUserAnimation(), {self.actionUser})
	end
end

function Battle:startActionProcess3()
	self.phase = "ActionProcess3"
	if self.action:hasAnimation() then
		self:addAnimation(self.action:getAnimation(), self.action:getTargets())
	end
	if self.action:hasCriticals() then
		gameDisplay:flash(1, 1, 1, 0.3, 8)
	end
	for _, v in ipairs(self.action:getResults()) do
		if v.target ~= self.actionUser then
			if v:isSuccess() then
				if v.alterHp and v.hpValue < 0 then
					v.target:performMotion("Damage")
				end
			else
				v.target:performMotion("Miss")
			end
		end
		self:addRefreshable(v.target)
	end
end

function Battle:startActionProcess4()
	self.phase = "ActionProcess4"
	self.action:applyResults()
	for _, v in ipairs(self.action:getMessages()) do
		self:addMessage("LongAuto", v)
	end
	self:executeActionDamage()
end

function Battle:startActionProcess5()
	self.phase = "ActionProcess5"
	if self.action:hasReturnMotion() then
		self.actionUser:performMotion("Return")
	end
end

function Battle:startActionProcess6()
	self.phase = "ActionProcess6"
	self:refreshMotions()
end

function Battle:startActionTerminate()
	self.phase = "ActionTerminate"
	self.actionUser:clearAction()
end

function Battle:startTurnPreTerminate()
	self.phase = "TurnPreTerminate"
	self:executeStatusDamage()
end

function Battle:startTurnTerminate()
	self.phase = "TurnTerminate"
	for _, v in ipairs(self.fighters) do
		v:removeStatusesOnTurnEnd()
		v:removeBuffsOnTurnEnd()
	end
	self:clearPreemptive()
	self:clearEscape()
	self:refreshMotions()
end

function Battle:startPreTerminate()
	self.phase = "PreTerminate"
	if self:isActorDefeated() then
		AM:fadeMusic(128)
		gameDisplay:color(0.5, 0.2, 0.1, 0.35, 128)
	elseif self:isEnemyDefeated() then
		AM:fadeMusic(64)
	elseif self:isEscapeSuccess() then
		-- TODO NOOP?
	end
end

function Battle:startTerminate()
	self.phase = "Terminate"
	if self:isActorDefeated() then
		self:executeActorDefeat()
	elseif self:isEnemyDefeated() then
		self:executeEnemyDefeat()
	elseif self:isEscapeSuccess() then
		self:executeEscape()
	end
end

---------------------------------------------------------------------------------------------------
-- Update methods
---------------------------------------------------------------------------------------------------

function Battle:updateBegin()
	self:startPostBegin()
end

function Battle:updatePostBegin()
	self:startInput()
end

function Battle:updateTurnBegin()
	if self:isOver() then
		self:startPreTerminate()
	else
		if #self.actionStack > 0 then
			self:startActionBegin()
		else
			self:startTurnPreTerminate()
		end
	end
end

function Battle:updateActionBegin()
	if self.action:isNone() or self.action:hasErrors() then
		self:startActionTerminate()
	else
		self:startActionProcess1()
	end
end

function Battle:updateActionProcess1()
	self:startActionProcess2()
end

function Battle:updateActionProcess2()
	self:startActionProcess3()
end

function Battle:updateActionProcess3()
	self:startActionProcess4()
end

function Battle:updateActionProcess4()
	self:startActionProcess5()
end

function Battle:updateActionProcess5()
	self:startActionProcess6()
end

function Battle:updateActionProcess6()
	self:startActionTerminate()
end

function Battle:updateActionTerminate()
	if self:isOver() then
		self:startPreTerminate()
	else
		if #self.actionStack > 0 then
			self:startActionBegin()
		else
			self:startTurnPreTerminate()
		end
	end
end

function Battle:updateTurnPreTerminate()
	self:startTurnTerminate()
end

function Battle:updateTurnTerminate()
	if self:isOver() then
		self:startPreTerminate()
	else
		self:startInput()
	end
end

function Battle:updatePreTerminate()
	self:startTerminate()
end

function Battle:updateTerminate()
	-- TODO
	SM:prevState()
end

function Battle:updatePhases()
	if self:isBusy() then
		return
	end
	local name = "update" .. self.phase
	if self[name] then
		self[name](self)
	end
end

function Battle:updateFighters()
	for _, v in ipairs(self.fighters) do
		v:update()
	end
end

function Battle:updateAnimations()
	for i = #self.animations, 1, -1 do
		if not self.animations[i]:isBusy() then
			table.remove(self.animations, i)
		end
	end
	for _, v in ipairs(self.animations) do
		v:update()
	end
end

function Battle:updateDamage()
	for i = #self.damage, 1, -1 do
		if not self.damage[i]:isBusy() then
			table.remove(self.damage, i)
		end
	end
	for _, v in ipairs(self.damage) do
		v:update()
	end
end

function Battle:update()
	self:updatePhases()
	self:updateFighters()
	self:updateAnimations()
	self:updateDamage()
end

return Battle