local Drawable = require "game.Drawable"

local Damage = Drawable:extend()

function Damage:initialize(type, value, x, y, delay)
	-- Shader
	self.shader = RM:getShader("default.glsl")
	-- Image/quads
	self.image = RM:getSystemImage("damage.png")
	self.quads = {}
	local dw = self:getDamageQuadW()
	local mw = self:getMissQuadW()
	local iw = dw * 10 + mw
	local ih = self:getQuadH()
	for i = 0, 9 do
		table.insert(self.quads, love.graphics.newQuad(i * dw, 0, dw, ih, iw, ih))
	end
	table.insert(self.quads, love.graphics.newQuad(dw * 10, 0, mw, ih, iw, ih))
	-- Motion
	self.motion = {}
	self.motionIndex = 1
	self.motionTimer = 0
	table.insert(self.motion, {code = "Wait", args = {delay or 0}})
	table.insert(self.motion, {code = "Show", args = {4}})
	table.insert(self.motion, {code = "Bounce", args = {32, 16}})
	table.insert(self.motion, {code = "Bounce", args = {16, 16}})
	table.insert(self.motion, {code = "Wait", args = {32}})
	table.insert(self.motion, {code = "Fade", args = {8}})
	table.insert(self.motion, {code = "Wait", args = {8}})
	table.insert(self.motion, {code = "End"})
	-- Type/value
	self.type = type
	self.value = value
	-- Position
	self.x = x
	self.y = y
	-- Bounce
	self.bh = 0
	self.bd = 0
	self.bc = 0
	-- Flash
	self.fi = 0
	self.fs = 0
	-- Opacity
	self.a = 0
	self.ta = 0
	self.sa = 0
end

function Damage:isBusy()
	return self.motionIndex <= #self.motion
end

function Damage:getDamageQuadW()
	return 16
end

function Damage:getMissQuadW()
	return 48
end

function Damage:getQuadH()
	return 24
end

function Damage:getBounceH()
	if self.bc < self.bd then
		return self.bc * (1 / self.bd - self.bc / (self.bd ^ 2)) * self.bh * 4
	end
	return 0
end

function Damage:getZ()
	return Damage.PRIORITY_DAMAGE
end

function Damage:show(duration)
	self.ta = 1
	if (duration or 0) > 0 then
		self.sa = (1 - self.a) / duration
	else
		self.a = 1
		self.sa = 0
	end
end

function Damage:fade(duration)
	self.ta = 0
	if (duration or 0) > 0 then
		self.sa = self.a / duration
	else
		self.a = 0
		self.sa = 0
	end
end

function Damage:bounce(height, duration)
	self.bh = height
	self.bd = duration
	self.bc = 0
end

function Damage:updateMotion(elem)
	local code = elem.code
	local args = elem.args or {}
	local name = code:sub(1, 1):lower() .. code:sub(2)
	if self[name] then
		self[name](self, unpack(args))
	else
		if code == "Wait" then
			if self.motionTimer > 0 then
				self.motionTimer = self.motionTimer - 1
			else
				self.motionTimer = args[1]
			end
			if self.motionTimer > 0 then
				self.motionIndex = self.motionIndex - 1
			end
		end
	end
end

function Damage:update()
	-- Motion
	if self.bc == self.bd then
		if self.motionIndex <= #self.motion then
			self:updateMotion(self.motion[self.motionIndex])
			self.motionIndex = self.motionIndex + 1
		end
	end
	-- Opacity
	if self.a < self.ta then
		self.a = math.min(self.ta, self.a + self.sa)
	end
	if self.a > self.ta then
		self.a = math.max(self.ta, self.a - self.sa)
	end
	-- Bounce
	if self.bc < self.bd then
		self.bc = self.bc + 1
	end
	-- Flash
	if self.fi > 0 then
		self.fi = math.max(0, self.fi - self.fs)
	else
		if self.a == 1 then
			self.fi = 0.5
			self.fs = self.fi / 8
		end
	end
end

function Damage:draw()
	local r = 0.5
	local g = 0.5
	local b = 0.5
	if self.type == "Hp" then
		if self.value > 0 then
			r = 0.25
			g = 0.5
			b = 0.25
		else
			r = 0.5
			g = 0.5
			b = 0.5
		end
	elseif self.type == "Mp" then
		if self.value > 0 then
			r = 0.25
			g = 0.5
			b = 0.5
		else
			r = 0.5
			g = 0.25
			b = 0.5
		end
	end
	self.shader:send("opacity", self.a)
	self.shader:sendColor("color", {r, g, b, 0.5})
	self.shader:sendColor("flash", {1, 1, 1, self.fi})
	self.shader:sendColor("displayColor", {0.5, 0.5, 0.5, 0.5})
	self.shader:sendColor("displayFlash", {1, 1, 1, 0})
	love.graphics.setShader(self.shader)
	love.graphics.setColor(1, 1, 1)
	if self.type == "Miss" then
		local w = self:getMissQuadW()
		local h = self:getQuadH()
		local x = self.x - w / 2
		local y = self.y - self:getBounceH() - h
		love.graphics.draw(self.image, self.quads[11], x, y)
	else
		local v = tostring(math.abs(self.value))
		local w = self:getDamageQuadW()
		local h = self:getQuadH()
		local x = self.x - v:len() * w / 2
		local y = self.y - self:getBounceH() - h
		for i = 1, v:len() do
			love.graphics.draw(self.image, self.quads[tonumber(v:sub(i, i)) + 1], x + (i - 1) * w, y)
		end
	end
end

return Damage