local Drawable = require "game.Drawable"

local Entity = Drawable:extend()

function Entity:initialize()
	-- Shader
	self.shader = RM:getShader("default.glsl")
	-- Image/quads
	self.image = nil
	self.quads = {}
	-- State
	self.visible = true
	-- Opacity
	self.a = 1
	self.ta = 1
	self.sa = 0
	-- Color
	self.r = 0.5
	self.g = 0.5
	self.b = 0.5
	self.s = 0.5
	self.tr = 0.5
	self.tg = 0.5
	self.tb = 0.5
	self.ts = 0.5
	self.sr = 0
	self.sg = 0
	self.sb = 0
	self.ss = 0
	-- Flash
	self.fr = 1
	self.fg = 1
	self.fb = 1
	self.fi = 0
	self.fs = 0
end

function Entity:isColoring()
	return self.r ~= self.tr or self.g ~= self.tg or self.b ~= self.tb or self.s ~= self.ts
end

function Entity:isFlashing()
	return self.fi > 0
end

function Entity:getColor()
	return self.r, self.g, self.b, self.s
end

function Entity:getFlash()
	return self.fr, self.fg, self.fb, self.fi
end

function Entity:getQuadIndex()
	return 0
end

function Entity:getQuadW()
	return (self.image and self.image:getWidth()) or 0
end

function Entity:getQuadH()
	return (self.image and self.image:getHeight()) or 0
end

function Entity:getW()
	return self:getQuadW()
end

function Entity:getH()
	return self:getQuadH()
end

function Entity:getX()
	return 0
end

function Entity:getY()
	return 0
end

function Entity:getYOffset()
	return 0
end

function Entity:setVisible(visible)
	self.visible = visible
end

function Entity:color(r, g, b, s, duration)
	self.tr = r
	self.tg = g
	self.tb = b
	self.ts = s
	if (duration or 0) > 0 then
		self.sr = math.abs(r - self.r) / duration
		self.sg = math.abs(g - self.g) / duration
		self.sb = math.abs(b - self.b) / duration
		self.ss = math.abs(s - self.s) / duration
	else
		self.r = r
		self.g = g
		self.b = b
		self.s = s
		self.sr = 0
		self.sg = 0
		self.sb = 0
		self.ss = 0
	end
end

function Entity:flash(r, g, b, intensity, duration)
	self.fr = r
	self.fg = g
	self.fb = b
	self.fi = intensity
	self.fs = intensity / duration
end

function Entity:show(duration)
	self.ta = 1
	if (duration or 0) > 0 then
		self.sa = (1 - self.a) / duration
	else
		self.a = 1
		self.sa = 0
	end
end

function Entity:fade(duration)
	self.ta = 0
	if (duration or 0) > 0 then
		self.sa = self.a / duration
	else
		self.a = 0
		self.sa = 0
	end
end

function Entity:update()
	if self.a < self.ta then
		self.a = math.min(self.ta, self.a + self.sa)
	end
	if self.a > self.ta then
		self.a = math.max(self.ta, self.a - self.sa)
	end
	if self.r < self.tr then
        self.r = math.min(self.tr, self.r + self.sr)
    end
    if self.r > self.tr then
        self.r = math.max(self.tr, self.r - self.sr)
    end
    if self.g < self.tg then
        self.g = math.min(self.tg, self.g + self.sg)
    end
    if self.g > self.tg then
        self.g = math.max(self.tg, self.g - self.sg)
    end
    if self.b < self.tb then
        self.b = math.min(self.tb, self.b + self.sb)
    end
    if self.b > self.tb then
        self.b = math.max(self.tb, self.b - self.sb)
    end
    if self.s < self.ts then
        self.s = math.min(self.ts, self.s + self.ss)
    end
    if self.s > self.ts then
        self.s = math.max(self.ts, self.s - self.ss)
    end
    if self.fi > 0 then
        self.fi = math.max(0, self.fi - self.fs)
	end
end

function Entity:draw()
	if not self.visible or not self.image then
		return
	end
	local i = self:getQuadIndex()
	local x = self:getX()
	local y = self:getY() - self:getYOffset()
	local ox = self:getQuadW() / 2
	local oy = self:getQuadH()
	self.shader:send("opacity", self.a)
	self.shader:sendColor("color", {self:getColor()})
	self.shader:sendColor("flash", {self:getFlash()})
	self.shader:sendColor("displayColor", {gameDisplay:getColor()})
	self.shader:sendColor("displayFlash", {gameDisplay:getFlash()})
	love.graphics.setShader(self.shader)
	love.graphics.setColor(1, 1, 1)
	if i > 0 then
		love.graphics.draw(self.image, self.quads[i], x, y, 0, 1, 1, ox, oy)
	else
		love.graphics.draw(self.image, x, y, 0, 1, 1, ox, oy)
	end
end

return Entity