local Object = require "core.Object"

local Target = Object:extend()

Target.TYPE_USER = 1
Target.TYPE_RANDOM_ALLY = 2
Target.TYPE_RANDOM_ALLY_ONLY = 3
Target.TYPE_RANDOM_OPPONENT = 4
Target.TYPE_RANDOM_OPPONENT_ONLY = 5
Target.TYPE_ALLY = 6
Target.TYPE_ALLY_ONLY = 7
Target.TYPE_OPPONENT = 8
Target.TYPE_OPPONENT_ONLY = 9
Target.TYPE_ALLIES = 10
Target.TYPE_ALLIES_ONLY = 11
Target.TYPE_OPPONENTS = 12
Target.TYPE_OPPONENTS_ONLY = 13

function Target:initialize()
	-- Image/quads
	self.image = RM:getSystemImage("target.png")
	self.quads = {}
	local qw = self:getQuadW()
	local qh = self:getQuadH()
	for i = 0, 1 do
		table.insert(self.quads, love.graphics.newQuad(i * qw, 0, qw, qh, qw * 2, qh))
	end
	-- Callbacks
	self.callbacks = {}
	-- State
	self.enabled = true
	self.all = false
	self.fixedGroup = false
	self.fixedIndex = false
	self.deadOnly = false
	self.actionCode = ""
	self.actionId = 0
	self.group = ""
	self.actorIndex = 1
	self.enemyIndex = 1
end

function Target:isSelectEnabled()
	return self.callbacks["Select"]
end

function Target:isCancelEnabled()
	return self.callbacks["Cancel"]
end

function Target:isAll()
	return self.all
end

function Target:isValidTarget(fighter)
	if fighter:isDead() then
		if fighter:isActor() then
			return self.deadOnly
		end
		if fighter:isEnemy() then
			return false
		end
	end
	return true
end

function Target:getActionCode()
	return self.actionCode
end

function Target:getActionId()
	return self.actionId
end

function Target:getGroup()
	return self.group
end

function Target:getIndex()
	if self.all then
		return 0
	end
	if self.group == "A" then
		return self.actorIndex
	end
	if self.group == "E" then
		return self.enemyIndex
	end
	return 0
end

function Target:getQuadW()
	return 32
end

function Target:getQuadH()
	return 48
end

function Target:setCallback(name, callback)
	self.callbacks[name] = callback
end

function Target:setEnabled(enabled)
	self.enabled = enabled
end

function Target:playCursorSound()
	AM:playCursorSound()
end

function Target:playSelectSound()
	AM:playSelectSound()
end

function Target:playCancelSound()
	AM:playCancelSound()
end

function Target:selectFirstActor()
	local i = self.actorIndex
	while not self:isValidTarget(gameActorGroup:getMember(self.actorIndex)) do
		self.actorIndex = self.actorIndex % gameActorGroup:getBattleSize() + 1
		if self.actorIndex == i then
			break
		end
	end
end

function Target:selectNextActor(wrap)
	local i = self.actorIndex
	while self.actorIndex + 1 <= gameActorGroup:getBattleSize() or wrap do
		self.actorIndex = self.actorIndex % gameActorGroup:getBattleSize() + 1
		if self.actorIndex == i or self:isValidTarget(gameActorGroup:getMember(self.actorIndex)) then
			break
		end
	end
	if self.actorIndex ~= i then
		self:playCursorSound()
	end
end

function Target:selectPrevActor(wrap)
	local i = self.actorIndex
	while self.actorIndex - 1 > 0 or wrap do
		self.actorIndex = (self.actorIndex - 1 + gameActorGroup:getBattleSize() - 1) % gameActorGroup:getBattleSize() + 1
		if self.actorIndex == i or self:isValidTarget(gameActorGroup:getMember(self.actorIndex)) then
			break
		end
	end
	if self.actorIndex ~= i then
		self:playCursorSound()
	end
end

function Target:selectFirstEnemy()
	local i = self.enemyIndex
	while not self:isValidTarget(gameEnemyGroup:getMember(self.enemyIndex)) do
		self.enemyIndex = self.enemyIndex % gameEnemyGroup:getBattleSize() + 1
		if self.enemyIndex == i then
			break
		end
	end
end

function Target:selectNextEnemy(wrap)
	local i = self.enemyIndex
	while self.enemyIndex + 1 <= gameEnemyGroup:getBattleSize() or wrap do
		self.enemyIndex = self.enemyIndex % gameEnemyGroup:getBattleSize() + 1
		if self.enemyIndex == i or self:isValidTarget(gameEnemyGroup:getMember(self.enemyIndex)) then
			break
		end
	end
	if self.enemyIndex ~= i then
		self:playCursorSound()
	end
end

function Target:selectPrevEnemy(wrap)
	local i = self.enemyIndex
	while self.enemyIndex - 1 > 0 or wrap do
		self.enemyIndex = (self.enemyIndex - 1 + gameEnemyGroup:getBattleSize() - 1) % gameEnemyGroup:getBattleSize() + 1
		if self.enemyIndex == i or self:isValidTarget(gameEnemyGroup:getMember(self.enemyIndex)) then
			break
		end
	end
	if self.enemyIndex ~= i then
		self:playCursorSound()
	end
end

function Target:select()
	self.callbacks["Select"]()
end

function Target:cancel()
	self.callbacks["Cancel"]()
end

function Target:setupTarget(code, id)
	self.all = false
	self.fixedGroup = false
	self.fixedIndex = false
	self.deadOnly = false
	self.actionCode = code
	self.actionId = id
	self.group = ""
	self.actorIndex = gameBattle:getIndex()
	self.enemyIndex = 1
	local data = ((code == "Skill" and dataSkills) or dataItems)[id]
	local type = data.targetType
	if type == Target.TYPE_USER then
		self.fixedGroup = true
		self.fixedIndex = true
		self.group = "A"
	end
	if type == Target.TYPE_RANDOM_ALLY then
		self.all = true
		self.fixedIndex = true
		self.group = "A"
	end
	if type == Target.TYPE_RANDOM_ALLY_ONLY then
		self.all = true
		self.fixedGroup = true
		self.fixedIndex = true
		self.group = "A"
	end
	if type == Target.TYPE_RANDOM_OPPONENT then
		self.all = true
		self.fixedIndex = true
		self.group = "E"
	end
	if type == Target.TYPE_RANDOM_OPPONENT_ONLY then
		self.all = true
		self.fixedGroup = true
		self.fixedIndex = true
		self.group = "E"
	end
	if type == Target.TYPE_ALLY then
		self.group = "A"
	end
	if type == Target.TYPE_ALLY_ONLY then
		self.fixedGroup = true
		self.group = "A"
	end
	if type == Target.TYPE_OPPONENT then
		self.group = "E"
	end
	if type == Target.TYPE_OPPONENT_ONLY then
		self.fixedGroup = true
		self.group = "E"
	end
	if type == Target.TYPE_ALLIES then
		self.all = true
		self.fixedIndex = true
		self.group = "A"
	end
	if type == Target.TYPE_ALLIES_ONLY then
		self.all = true
		self.fixedGroup = true
		self.fixedIndex = true
		self.group = "A"
	end
	if type == Target.TYPE_OPPONENTS then
		self.all = true
		self.fixedIndex = true
		self.group = "E"
	end
	if type == Target.TYPE_OPPONENTS_ONLY then
		self.all = true
		self.fixedGroup = true
		self.fixedIndex = true
		self.group = "E"
	end
	self.deadOnly = data.deadOnly
	self:selectFirstActor()
	self:selectFirstEnemy()
end

function Target:update()
	if not self.enabled then
		return
	end
	if IM:isRepeated("Up") then
		if not self.fixedIndex then
			local wrap = IM:isPressed("Up")
			if self.group == "A" then
				self:selectPrevActor(wrap)
			end
			if self.group == "E" then
				self:selectPrevEnemy(wrap)
			end
		end
	elseif IM:isRepeated("Right") then
		if not self.fixedGroup then
			if self.group == "E" then
				self.group = "A"
				self:playCursorSound()
			end
		end
	elseif IM:isRepeated("Down") then
		if not self.fixedIndex then
			local wrap = IM:isPressed("Down")
			if self.group == "A" then
				self:selectNextActor(wrap)
			end
			if self.group == "E" then
				self:selectNextEnemy(wrap)
			end
		end
	elseif IM:isRepeated("Left") then
		if not self.fixedGroup then
			if self.group == "A" then
				self.group = "E"
				self:playCursorSound()
			end
		end
	elseif IM:isPressed("Select") then
		if self:isSelectEnabled() then
			self:select()
			self:playSelectSound()
			IM:clear()
		else
			self:playCancelSound()
		end
	elseif IM:isPressed("Cancel") then
		if self:isCancelEnabled() then
			self:cancel()
			self:playCancelSound()
			IM:clear()
		end
	end
end

function Target:drawTarget(fighter)
	local i = 0
	local w = self:getQuadW()
	local h = self:getQuadH()
	local x = fighter:getX()
	local y = fighter:getY() - (fighter:getH() + h) / 2
	if fighter:isActor() then
		i = 1
		x = x - fighter:getW() / 2 - w
	end
	if fighter:isEnemy() then
		i = 2
		x = x + fighter:getW() / 2
	end
	love.graphics.draw(self.image, self.quads[i], x, y)
end

function Target:draw()
	if not self.enabled then
		return
	end
	love.graphics.setShader()
	love.graphics.setColor(1, 1, 1, self.all and 0.5 or 1)
	if self.group == "A" then
		for i, v in ipairs(gameActorGroup:getBattleMembers()) do
			if i == self.actorIndex or self.all and self:isValidTarget(v) then
				self:drawTarget(v)
			end
		end
	end
	if self.group == "E" then
		for i, v in ipairs(gameEnemyGroup:getBattleMembers()) do
			if i == self.enemyIndex or self.all and self:isValidTarget(v) then
				self:drawTarget(v)
			end
		end
	end
end

return Target