local State = require "state.State"
local Action = require "game.Action"
local Indicator = require "game.Indicator"
local Target = require "game.Target"
local BattleGroupWindow = require "window.BattleGroupWindow"
local ActionWindow = require "window.ActionWindow"
local InlineActionWindow = require "window.InlineActionWindow"
local SkillWindow = require "window.SkillWindow"
local ItemWindow = require "window.ItemWindow"
local ClosableInfoWindow = require "window.ClosableInfoWindow"
local AutoclosableInfoWindow = require "window.AutoclosableInfoWindow"

local BattleState = State:extend()

function BattleState:initialize()
	State.initialize(self)
	-- Indicator
	self.indicator = Indicator:new()
	self.indicator:setEnabled(false)
	-- Target
	self.target = Target:new()
	self.target:setEnabled(false)
	self.target:setCallback("Select", function() self:targetSelect() end)
	self.target:setCallback("Cancel", function() self:targetCancel() end)
	-- Group window
	self.groupWindow = BattleGroupWindow:new(32, 320, 576, 160)
	self.groupWindow:setEnabled(false)
	self.groupWindow:setCursorVisible(false)
	-- Group input window
	self.groupInputWindow = InlineActionWindow:new(120, 0, 400, 64, {
		{id = "Fight", name = TM.BATTLE_FIGHT},
		{id = "Escape", name = TM.BATTLE_ESCAPE}
	})
	self.groupInputWindow:setCallback("Fight", function() self:groupInputFight() end)
	self.groupInputWindow:setCallback("Escape", function() self:groupInputEscape() end)
	self.groupInputWindow:setCallback("Closed", function() self:groupInputClosed() end)
	-- Actor input window
	self.actorInputWindow = ActionWindow:new(40, 320, 160, 160)
	self.actorInputWindow:setCallback("Attack", function() self:actorInputAttack() end)
	self.actorInputWindow:setCallback("Defend", function() self:actorInputDefend() end)
	self.actorInputWindow:setCallback("Magic", function() self:actorInputSkill() end)
	self.actorInputWindow:setCallback("Technique", function() self:actorInputSkill() end)
	self.actorInputWindow:setCallback("Summon", function() self:actorInputSkill() end)
	self.actorInputWindow:setCallback("Item", function() self:actorInputItem() end)
	self.actorInputWindow:setCallback("Cancel", function() self:actorInputCancel() end)
	self.actorInputWindow:setCallback("Closed", function() self:actorInputClosed() end)
	-- Skill window
	self.skillWindow = SkillWindow:new(48, 320, 576, 160)
	self.skillWindow:setCallback("Select", function() self:skillSelect() end)
	self.skillWindow:setCallback("Cancel", function() self:skillCancel() end)
	-- Item window
	self.itemWindow = ItemWindow:new(48, 320, 576, 160, gameActorGroup:getItems())
	self.itemWindow:setCallback("Select", function() self:itemSelect() end)
	self.itemWindow:setCallback("Cancel", function() self:itemCancel() end)
	-- Message windows
	self.messageWindow1 = AutoclosableInfoWindow:new(192, 0, 256, 64, "", AutoclosableInfoWindow.ALIGN_CENTER)
	self.messageWindow1:setCallback("Closed", function() self:messageClosed() end)
	self.messageWindow2 = ClosableInfoWindow:new(32, 0, 576, 64)
	self.messageWindow2:setCallback("Closed", function() self:messageClosed() end)
	self.messageWindow3 = AutoclosableInfoWindow:new(32, 0, 576, 64)
	self.messageWindow3:setCallback("Closed", function() self:messageClosed() end)
end

---------------------------------------------------------------------------------------------------
-- Callbacks
---------------------------------------------------------------------------------------------------

function BattleState:targetSelect()
	-- Action
	local index = gameBattle:getIndex()
	local actor = gameActorGroup:getMember(index)
	actor:setAction(Action:new(self.target:getActionCode(), self.target:getActionId(), actor, self.target:isAll(), self.target:getGroup(), self.target:getIndex()))
	-- Target/indicator
	self.target:setEnabled(false)
	self.indicator:setEnabled(false)
	-- Windows
	self.groupWindow:setCursorVisible(false)
	self.actorInputWindow:setEnabled(true)
	self.actorInputWindow:close()
	self.skillWindow:setEnabled(true)
	self.skillWindow:close()
	self.itemWindow:setEnabled(true)
	self.itemWindow:close()
	-- Actor
	gameBattle:selectNextActor()
end

function BattleState:targetCancel()
	self.target:setEnabled(false)
	if self.actorInputWindow:getActionId() == "Attack" then
		self.actorInputWindow:setEnabled(true)
	end
	if self.actorInputWindow:getActionId() == "Technique" or self.actorInputWindow:getActionId() == "Magic" or self.actorInputWindow:getActionId() == "Summon" then
		self.skillWindow:setEnabled(true)
	end
	if self.actorInputWindow:getActionId() == "Item" then
		self.itemWindow:setEnabled(true)
	end
end

function BattleState:groupInputFight()
	self.groupInputWindow:close()
	gameBattle:selectNextActor()
end

function BattleState:groupInputEscape()
	self.groupInputWindow:close()
	gameBattle:selectEscape()
end

function BattleState:groupInputClosed()
	if not gameBattle:isActorInputPhase() then
		gameBattle:startTurnBegin()
	end
end

function BattleState:actorInputAttack()
	self.target:setEnabled(true)
	self.target:setupTarget("Skill", Action.SKILL_ATTACK_ID)
	self.actorInputWindow:setEnabled(false)
end

function BattleState:actorInputDefend()
	local actor = gameActorGroup:getMember(gameBattle:getIndex())
	actor:setAction(Action:new("Skill", Action.SKILL_DEFEND_ID, actor, false, "A", gameBattle:getIndex()))
	self.indicator:setEnabled(false)
	self.groupWindow:setCursorVisible(false)
	self.actorInputWindow:close()
	gameBattle:selectNextActor()
end

function BattleState:actorInputSkill()
	local actor = gameActorGroup:getMember(gameBattle:getIndex())
	self.actorInputWindow:setEnabled(false)
	if self.actorInputWindow:getActionId() == "Technique" then
		self.skillWindow:setList(actor:getTechniques())
	end
	if self.actorInputWindow:getActionId() == "Magic" then
		self.skillWindow:setList(actor:getMagic())
	end
	if self.actorInputWindow:getActionId() == "Summon" then
		self.skillWindow:setList(actor:getSummon())
	end
	self.skillWindow:setIndex(1)
	self.skillWindow:open()
end

function BattleState:actorInputItem()
	self.actorInputWindow:setEnabled(false)
	self.itemWindow:setIndex(1)
	self.itemWindow:open()
end

function BattleState:actorInputCancel()
	self.indicator:setEnabled(false)
	self.groupWindow:setCursorVisible(false)
	self.actorInputWindow:close()
	gameBattle:selectPrevActor()
end

function BattleState:actorInputClosed()
	if not gameBattle:isGroupInputPhase() and not gameBattle:isActorInputPhase() then
		gameBattle:startTurnBegin()
	end
end

function BattleState:skillSelect()
	self.target:setEnabled(true)
	self.target:setupTarget("Skill", self.skillWindow:getSkillId())
	self.skillWindow:setEnabled(false)
end

function BattleState:skillCancel()
	self.actorInputWindow:setEnabled(true)
	self.skillWindow:close()
end

function BattleState:itemSelect()
	self.target:setEnabled(true)
	self.target:setupTarget("Item", self.itemWindow:getItemId())
	self.itemWindow:setEnabled(false)
end

function BattleState:itemCancel()
	self.actorInputWindow:setEnabled(true)
	self.itemWindow:close()
end

function BattleState:messageClosed()
	gameBattle:popMessage()
end

---------------------------------------------------------------------------------------------------
-- Setup methods
---------------------------------------------------------------------------------------------------

function BattleState:setupActorInputWindow()
	local actor = gameActorGroup:getMember(gameBattle:getIndex())
	local actions = {}
	-- Attack/defend
	table.insert(actions, {id = "Attack", name = TM.BATTLE_ATTACK})
	table.insert(actions, {id = "Defend", name = TM.BATTLE_DEFEND})
	-- Skills
	if actor:hasTechniques() then
		table.insert(actions, {id = "Technique", name = TM.BATTLE_TECHNIQUE})
	end
	if actor:hasMagic() then
		table.insert(actions, {id = "Magic", name = TM.BATTLE_MAGIC})
	end
	if actor:hasSummon() then
		table.insert(actions, {id = "Summon", name = TM.BATTLE_SUMMON})
	end
	-- Item
	table.insert(actions, {id = "Item", name = TM.BATTLE_ITEM})
	-- Actions
	self.actorInputWindow:setActions(actions)
end

---------------------------------------------------------------------------------------------------
-- State start/stop
---------------------------------------------------------------------------------------------------

function BattleState:start()
	gameTransition:fadeIn(32)
	gameBattle:startBegin()
end

function BattleState:stop()
	gameTransition:fadeOut(32)
	AM:fadeMusic(32)
end

---------------------------------------------------------------------------------------------------
-- Update
---------------------------------------------------------------------------------------------------

function BattleState:updateObjects()
	gameDisplay:update()
	gameBattle:update()
	self.indicator:update()
	self.target:update()
end

function BattleState:updateWindows()
	self.groupWindow:update()
	self.groupInputWindow:update()
	self.actorInputWindow:update()
	self.skillWindow:update()
	self.itemWindow:update()
	self.messageWindow1:update()
	self.messageWindow2:update()
	self.messageWindow3:update()
end

function BattleState:updateVisibility()
	-- Group window
	-- TODO
	if not SM:isStateChanging() and not gameTransition:isBusy() then
		if self.groupWindow:isClosed() then
			self.groupWindow:open()
		end
	end
	-- TODO
	-- Group input window
	if gameBattle:isGroupInputPhase() then
		if self.groupInputWindow:isClosed() and self.actorInputWindow:isClosed() then
			self.groupInputWindow:setIndex(1)
			self.groupInputWindow:open()
		end
	end
	-- Actor input window
	if gameBattle:isActorInputPhase() then
		if self.groupInputWindow:isClosed() and self.actorInputWindow:isClosed() then
			local index = gameBattle:getIndex()
			-- Indicator
			self.indicator:setEnabled(true)
			self.indicator:setIndex(index)
			-- Group window
			self.groupWindow:setCursorVisible(true)
			self.groupWindow:setIndex(index)
			-- Input window
			self:setupActorInputWindow()
			self.actorInputWindow:setIndex(1)
			self.actorInputWindow:open()
			-- Sound
			AM:playActionSound()
		end
	end
	-- Message windows
	if gameBattle:hasMessages() then
		local message = gameBattle:getMessage()
		if message.type == "Auto" then
			if self.messageWindow1:isClosed() then
				self.messageWindow1:setText(message.text)
				self.messageWindow1:open()
			end
		end
		if message.type == "Long" then
			if self.messageWindow2:isClosed() then
				self.messageWindow2:setText(message.text)
				self.messageWindow2:open()
			end
		end
		if message.type == "LongAuto" then
			if self.messageWindow3:isClosed() then
				self.messageWindow3:setText(message.text)
				self.messageWindow3:open()
			end
		end
	end
end

function BattleState:update()
	State.update(self)
	self:updateObjects()
	self:updateWindows()
	self:updateVisibility()
end

---------------------------------------------------------------------------------------------------
-- Draw
---------------------------------------------------------------------------------------------------

function BattleState:drawBackground()
	local shader = RM:getShader("default.glsl")
	shader:send("opacity", 1)
	shader:sendColor("color", {0.5, 0.5, 0.5, 0.5})
	shader:sendColor("flash", {1, 1, 1, 0})
	shader:sendColor("displayColor", {gameDisplay:getColor()})
	shader:sendColor("displayFlash", {gameDisplay:getFlash()})
	-- TODO
	love.graphics.setShader(shader)
	love.graphics.setColor(1, 1, 1)
	love.graphics.draw(RM:getBackgroundImage("rnhgrass.png"))
	-- TODO
end

function BattleState:drawObjects()
	-- Drawables
	local drawables = {}
	for _, v in ipairs(gameBattle:getFighters()) do
		table.insert(drawables, v)
	end
	for _, v in ipairs(gameBattle:getAnimations()) do
		table.insert(drawables, v)
	end
	for _, v in ipairs(gameBattle:getDamage()) do
		table.insert(drawables, v)
	end
	table.sort(drawables)
	for _, v in ipairs(drawables) do
		v:draw()
	end
	-- Indicator
	self.indicator:draw()
	-- Target
	self.target:draw()
end

function BattleState:drawWindows()
	self.groupWindow:draw()
	self.groupInputWindow:draw()
	self.actorInputWindow:draw()
	self.skillWindow:draw()
	self.itemWindow:draw()
	self.messageWindow1:draw()
	self.messageWindow2:draw()
	self.messageWindow3:draw()
end

function BattleState:draw()
	self:drawBackground()
	self:drawObjects()
	self:drawWindows()
	State.draw(self)
end

return BattleState