local Window = require "window.Window"

local OptionWindow = Window:extend()

function OptionWindow:initialize(x, y, w, h)
	Window.initialize(self, x, y, w, h)
	self.cursorVisible = true
	self.index = 1
end

function OptionWindow:isCursorMovable()
	return true
end

function OptionWindow:isSelectEnabled()
	return self:hasCallback("Select")
end

function OptionWindow:isCancelEnabled()
	return self:hasCallback("Cancel")
end

function OptionWindow:isItemEnabled(index)
	return true
end

function OptionWindow:getSize()
	return 0
end

function OptionWindow:getCols()
	return 1
end

function OptionWindow:getRows()
	return math.ceil(self:getSize() / self:getCols())
end

function OptionWindow:getVisibleRows()
	return math.ceil(self:getContentH() / self:getItemH())
end

function OptionWindow:getTopRow()
	return self.oy / self:getItemH() + 1
end

function OptionWindow:getBottomRow()
	return self:getTopRow() + self:getVisibleRows() - 1
end

function OptionWindow:getItemH()
	return self:getLineH()
end

function OptionWindow:getItemHPad()
	return self:getDefaultSpacing()
end

function OptionWindow:getItemRect(index)
	local w = self:getContentW() / self:getCols()
	local h = self:getItemH()
	local x = ((index - 1) % self:getCols()) * w
	local y = math.floor((index - 1) / self:getCols()) * h
	return x, y, w, h
end

function OptionWindow:getIndex()
	return self.index
end

function OptionWindow:setCursorVisible(cursorVisible)
	self.cursorVisible = cursorVisible
end

function OptionWindow:setTopRow(row)
	self.oy = (row - 1) * self:getItemH()
end

function OptionWindow:setBottomRow(row)
	self:setTopRow(row - self:getVisibleRows() + 1)
end

function OptionWindow:setIndex(index)
	self.index = index
	local row = math.floor((index - 1) / self:getCols()) + 1
	if row < self:getTopRow() then
		self:setTopRow(row)
	end
	if row > self:getBottomRow() then
		self:setBottomRow(row)
	end
	if self:hasCallback("IndexChange") then
		self:invokeCallback("IndexChange")
	end
end

function OptionWindow:moveCursor(direction, wrap)
	if self:getSize() == 0 then
		return
	end
	local oi = self.index
	if direction == 0 then
		if self:getRows() > 1 and (self.index - self:getCols() > 0 or wrap) then
			self:setIndex((self:getIndex() - self:getCols() + self:getSize() - 1) % self:getSize() + 1)
		end
	end
	if direction == 1 then
		if self:getCols() > 1 and (self.index + 1 <= self:getSize() or wrap) then
			self:setIndex((self:getIndex() + 1 - 1) % self:getSize() + 1)
		end
	end
	if direction == 2 then
		if self:getRows() > 1 and (self.index + self:getCols() <= self:getSize() or wrap) then
			self:setIndex((self:getIndex() + self:getCols() - 1) % self:getSize() + 1)
		end
	end
	if direction == 3 then
		if self:getCols() > 1 and (self.index - 1 > 0 or wrap) then
			self:setIndex((self:getIndex() - 1 + self:getSize() - 1) % self:getSize() + 1)
		end
	end
	if self.index ~= oi then
		self:playCursorSound()
	end
end

function OptionWindow:select()
	self:invokeCallback("Select")
end

function OptionWindow:cancel()
	self:invokeCallback("Cancel")
end

function OptionWindow:update()
	Window.update(self)
	-- Input
	if self.enabled and self:isOpen() and self:isCursorMovable() then
		if IM:isRepeated("Up") then
			self:moveCursor(0, IM:isPressed("Up"))
		elseif IM:isRepeated("Right") then
			self:moveCursor(1, IM:isPressed("Right"))
		elseif IM:isRepeated("Down") then
			self:moveCursor(2, IM:isPressed("Down"))
		elseif IM:isRepeated("Left") then
			self:moveCursor(3, IM:isPressed("Left"))
		elseif IM:isPressed("Select") then
			if self:getSize() > 0 and self:isSelectEnabled() and self:isItemEnabled(self.index) then
				self:select()
				self:playSelectSound()
				IM:clear()
			else
				self:playCancelSound()
			end
		elseif IM:isPressed("Cancel") then
			if self:isCancelEnabled() then
				self:cancel()
				self:playCancelSound()
				IM:clear()
			end
		end
	end
end

function OptionWindow:drawItem(index, x, y, w, h)
end

function OptionWindow:drawInterface()
	if self:getTopRow() > 1 then
		self:drawTopArrow()
	end
	if self:getBottomRow() < self:getRows() then
		self:drawBottomArrow()
	end
end

function OptionWindow:drawContent()
	-- Cursor
	if self.cursorVisible then
		local x, y, w, h = self:getItemRect(self.index)
		self:drawCursor(x, y, w, h, self.enabled and not self.blink)
	end
	-- Items
	for i = 1, self:getSize() do
		local x, y, w, h = self:getItemRect(i)
		x = x + self:getItemHPad()
		w = w - self:getItemHPad() * 2
		self:drawItem(i, x, y, w, h)
	end
end

return OptionWindow